<?php
/**
 * @version		$Id: jailserver_manager.class.php,v 1.6 2009/11/04 09:51:15 juanca Exp $
 * @package		vpl. vpl jailservers manager class definition
 * @copyright	Copyright (C) 2009 Juan Carlos Rodríguez-del-Pino. All rights reserved.
 * @license		GNU/GPL, see LICENSE.txt or http://www.gnu.org/licenses/gpl-2.0.html
 * @author		Juan Carlos Rodriguez-del-Pino
 **/

/**
 * vpl_jailserver_manager is a utility class to manage
 * the jail servers. get_Server if the key function
 *
 */
class vpl_jailserver_manager{
	const recheck=300; //300 sec = 5 min ¿optional setable?
	const table='vpl_jailservers';

	/**
	 * Check if the server is tagged as down
	 * @param url $server
	 * @return boolean
	 */
	static private function is_checkable($server){
		$info= get_record('vpl_jailservers','server',$server);
		if($info != null){
			if($info->lastfail+self::recheck > time()){
				return false;
			}
		}
		return true;
	}

	/**
	 * Tag the server as down
	 * @param URL $server
	 * @param string $strerror
	 * @return void
	 */
	static private function server_fail($server,$strerror){
		if($strerror==null){
			$strerror='';
		}
		$info= get_record(self::table,'server',$server);
		if($info != null){
			$info->lastfail=time();
			$info->laststrerror=$strerror;
			$info->nfails++;
			update_record(self::table,$info);
		}else{
			$info = new stdClass();
			$info->server = $server;
			$info->lastfail=time();
			$info->laststrerror=$strerror;
			$info->nfails=1;
			insert_record(self::table,$info);
		}
	}

	/**
	 * Return a valid server to be used, May tag some servers as faulty
	 * @param int $maxmemory requiered
	 * @return URL
	 */
	static function get_server($maxmemory,$checkall=false){
		global $CFG;
		$data = new stdClass();
		$data->maxmemory=$maxmemory;
		$nl = vpl_detect_newline($CFG->vpl_jail_servers);
		$serverlist = explode("\n",$CFG->vpl_jail_servers);
		$requestReady = xmlrpc_encode_request('status',$data);
		shuffle($serverlist);
		$http=null;
		foreach ($serverlist as $server) {
			$server = trim($server);
			if($server>'' && $server[0]!='#' &&
				($checkall || vpl_jailserver_manager::is_checkable($server))){
				$http = new vpl_HTTP_request($requestReady);
				if($http->try_server($server)){
					$raw_response=$http->get_response();
					$response = xmlrpc_decode($raw_response);
					if(xmlrpc_is_fault ($response)){
						p(print_r($response));
					}
					$status=$response['status'];
					if($status == 'ready'){
						return $server;
					}
					else if($status!='busy'){
						//TODO need actions?
						//vpl_jailserver_manager::server_fail($server,'Alive but not ready :'.$raw_response);
						continue;
					}
				}
				vpl_jailserver_manager::server_fail($server,$http->get_error());
			}
		}
		if(!$checkall){
			return self::get_server($maxmemory,true);
		}
		return false;
	}

	/**
	 * Clear servers table and check for every one again
	 * @return string server list with info
	 */
	static function reset_servers(){
		global $CFG;
		delete_records('vpl_jailservers');
		$nl = vpl_detect_newline($CFG->vpl_jail_servers);
		$serverlist = explode($nl,$CFG->vpl_jail_servers);
		$requestReady = xmlrpc_encode_request('status',null);
		$result='';
		$http=null;
		foreach ($serverlist as $server) {
			$server = trim($server);
			if($server>'' && vpl_jailserver_manager::is_checkable($server)){
				$http = new vpl_HTTP_request($requestReady);
				if($http->try_server($server)){
					$response = xmlrpc_decode($http->get_response());
					//TODO use new format
					if($response == 'yes'){
						$result .=$server+" (OK)\n";
					}
					else{
						$result .=$server+' (Alive but not ready :'.$http->get_response().")\n";
					}
				}else{
					vpl_jailserver_manager::server_fail($server,$http->get_error());
					$result .=$server+' (Not ready :'.$http->get_error().")\n";
				}
			}
		}
		return $result;
	}
}

?>