<?php
/**
 * @version		$Id: diff.php,v 1.4 2010/01/19 17:58:04 juanca Exp $
 * @package		vpl. List most similar submission files
 * @copyright	Copyright (C) 2009 Juan Carlos Rodríguez-del-Pino. All rights reserved.
 * @license		GNU/GPL, see LICENSE.txt or http://www.gnu.org/licenses/gpl-2.0.html
 * @author		Juan Carlos Rodriguez-del-Pino
 **/

require_once dirname(__FILE__).'/../../../config.php';
require_once dirname(__FILE__).'/../locallib.php';
require_once dirname(__FILE__).'/../vpl.class.php';
require_once dirname(__FILE__).'/../vpl_submission.class.php';
require_once dirname(__FILE__).'/../views/sh_base.class.php';
require_once dirname(__FILE__).'/similarity_factory.class.php';
require_once dirname(__FILE__).'/similarity_base.class.php';
class vpl_diff{
	static function removeAlphaNum($line){
		$ret='';
		$l= strlen($line);
		//Parse line to remove alphanum chars
		for($i=0; $i<$l; $i++){
			$c=$line[$i];
			if(($c > ' ' && $c < '0') || ($c > 'Z' && $c < 'a') || ($c > 'z' && $c <= '~')){
				$ret.=$c;
			}
		}
		return $ret;
	}

	static function diffLine($line1,$line2){
		//Compare plain text
		if($line1==$line2){
			return 6;
		}
		//Compare trimed text
		$line1=trim($line1);
		$line2=trim($line2);
		if($line1==$line2){
			return 5;
		}
		//Compare filtered text (removing alphanum)
		$line1=self::removeAlphaNum($line1);
		$line2=self::removeAlphaNum($line2);
		if($line1==$line2){
			return 4;
		}
		//Compare start of line
		$l=4;
		if($l>strlen($line1)){
			$l=strlen($line1);
		}
		if($l>strlen($line2)){
			$l=strlen($line2);
		}
		for($i=0; $i<$l; ++$i){
			if($line1[$i] !=$line2[$i]){
				break;
			}
		}
		return $i-1;
	}

	static function newLineInfo($type,$ln1,$ln2=0){
		$ret = new StdClass;
		$ret->type = $type;
		$ret->ln1 = $ln1;
		$ret->ln2 = $ln2;
		return $ret;
	}

	static function initAuxiliarMatrices(&$matrix,&$prev,$nl1,$nl2){
		// matrix[0..nl1+1][0..nl2+1]=0
		$row = array_pad(array(),$nl2+1,0);
		$matrix = array_pad(array(),$nl1+1,$row);
		// prev[0..nl1+1][0..nl2+1]=0
		$row = array_pad(array(),$nl2+1,0);
		$prev = array_pad(array(),$nl1+1,$row);

		//update first row and column
		for($i=0; $i<=$nl1;$i++){
			$matriz[$i][0]=0;
			$prev[$i][0]=-1;
			$value->i=$i;
		}
		$value->i=0;
		for($j=0; $j<=$nl2;$j++){
			$matriz[0][$j]=0;
			$prev[0][$j]=1;
		}
	}
	
	static function calculateDiff($lines1,$lines2){
		$ret = array();
		$nl1=count($lines1);
		$nl2=count($lines2);
		if($nl1==0 && $nl2==0){
			return false;
		}
		if($nl1==0){ //There is no first file
			foreach($lines2 as $pos => $line){
				$ret[] = self::newLineInfo('>',0,$pos+1);
			}
			return $ret;
		}
		if($nl2==0){ //There is no second file
			foreach($lines1 as $pos => $line){
				$ret[] = self::newLineInfo('<',$pos+1);
			}
			return $ret;
		}
		self::initAuxiliarMatrices($matrix,$prev,$nl1,$nl2);

		//Matrix processing
		for($i=1; $i <= $nl1;$i++){
			$line=$lines1[$i-1];
			for($j=1; $j<=$nl2;$j++){
				$punishment=1;
				$prize=self::diffLine($line,$lines2[$j-1]);
				$best = 0;
				$max=$matrix[$i-1][$j-1]+$prize;
				if($matrix[$i][$j-1]-$punishment>$max) {
					$max=$matrix[$i][$j-1]-$punishment;
					$best = 1;
				}
				if($matrix[$i-1][$j]-$punishment>$max){
					$max=$matrix[$i-1][$j]-$punishment;
					$best = -1;
				}
				$matrix[$i][$j]=$max;
				$prev[$i][$j]=$best;
			}
		}

		//Calculate show info
		$limit=$nl1+$nl2+2;
		$pairs = array();
		$pi=$nl1;
		$pj=$nl2;
		while((!($pi == 0 && $pj == 0)) && $limit>0){
			$pair = new stdClass();
			$pair->i = $pi;
			$pair->j = $pj;
			$pairs[]=$pair;
			$p = $prev[$pi][$pj];
			if($p == 0){
				$pi--;
				$pj--;
			}elseif($p == -1){
				$pi--;
			}else{
				$pj--;
			}
			$limit--;
		}

		krsort($pairs);

		$prevpair->i=0;
		$prevpair->j=0;
		foreach($pairs as $pair){
			if($pair->i == $prevpair->i+1 && $pair->j == $prevpair->j+1){ //Regular advance
				if($lines1[$pair->i-1] == $lines2[$pair->j-1]){ //Equeals
					$ret[]=self::newLineInfo('=',$pair->i,$pair->j);
				}else{
					$ret[]=self::newLineInfo('#',$pair->i,$pair->j);
				}
			}elseif($pair->i == $prevpair->i+1){ //Removed next line
				$ret[]=self::newLineInfo('<',$pair->i);
			}elseif($pair->j == $prevpair->j+1){ //Added one line
				$ret[]=self::newLineInfo('>',0,$pair->j);
			}else{
				debugging("Internal error ".print_r($pair,true)." ".print_r($prevpair,true));
			}
			$prevpair=$pair;
		}
		return $ret;
	}

	
	static function show($filename1, $data1, $HTMLheader1, $filename2, $data2, $HTMLheader2) {
		//Get file lines
		$nl = vpl_detect_newline($data1);
		$lines1 = explode($nl,$data1);
		$nl = vpl_detect_newline($data2);
		$lines2 = explode($nl,$data2);
		//Get dif as an array of info
		$diff = self::calculateDiff($lines1,$lines2);
		if($diff === false){
			return;
		}
		$separator= array('<'=>' <<< ', '>'=>' >>> ','='=>' === ','#'=>' ### ');
		$emptyline="&nbsp;\n";
		$data1='';
		$data2='';
		$datal1='';
		$datal2='';
		$diffl='';
		$lines1[-1]='';
		$lines2[-1]='';
		foreach($diff as $line){
			$diffl.= $separator[$line->type]."\n";
			if($line->ln1){
				$datal1.=$line->ln1." \n";
				$data1.=$lines1[$line->ln1-1]."\n";
			}else{
				$data1.="\n";
				$datal1.=$emptyline;
			}
			if($line->ln2){
				$datal2.=$line->ln2." \n";
				$data2.=$lines2[$line->ln2-1]."\n";
			}else{
				$data2.="\n";
				$datal2.=$emptyline;
			}
		}
		echo '<div style="width: 900px">';
		//Header
		echo '<div style="float:left; width: 445px">';
		echo $HTMLheader1;
		echo '</div>';
		echo '<div style="float:left; width: 445px">';
		echo $HTMLheader2;
		echo '</div>';
		echo '<div style="clear:both;"></div>';
		//Files
		echo '<div style="float:left; text-align: right; width: 40px">';
		echo '<pre class="'.vpl_sh_base::c_general.'">';
		echo $datal1;
		echo '</pre>';
		echo '</div>';
		echo '<div style="float:left; width: 390px; overflow:auto">';
		$shower= vpl_sh_factory::get_sh($filename1);
		$shower->print_file($filename1,$data1,false);
		echo '</div>';
		echo '<div style="float:left; width: 40px">';
		echo '<pre class="'.vpl_sh_base::c_general.'">';
		echo $diffl;
		echo '</pre>';
		echo '</div>';
		echo '<div style="float:left; text-align: right; width: 40px">';
		echo '<pre class="'.vpl_sh_base::c_general.'">';
		echo $datal2;
		echo '</pre>';
		echo '</div>';
		echo '<div style="float:left; width: 390px; overflow:auto">';
		$shower= vpl_sh_factory::get_sh($filename2);
		$shower->print_file($filename2,$data2,false);
		echo '</div>';
		echo '</div>';
		echo '<div style="clear:both;"></div>';
		
	}

	static function vpl_get_similfile($f,&$HTMLheader,&$filename,&$data){
		$HTMLheader='';
		$filename='';
		$data='';
		$type = required_param('type'.$f, PARAM_INT);
		if($type==1){
			$subid = required_param('subid'.$f, PARAM_INT);
			$filename = required_param('filename'.$f, PARAM_CLEANFILE);
			$subinstance = get_record('vpl_submissions','id',$subid);
			if($subinstance !== false){
				$vpl = new mod_vpl(false,$subinstance->vpl);
				$vpl->require_capability(VPL_SIMILARITY_CAPABILITY);
				$submission = new mod_vpl_submission($vpl,$subinstance);
				$user = get_record('user','id',$subinstance->userid);
				if($user){
					$HTMLheader .='<a href="'.vpl_mod_href('/forms/submissionview.php',
								'id',$vpl->get_course_module()->id,'userid',$subinstance->userid).'">';
				}
				$HTMLheader .=s($filename).' ';
				if($user){
					$HTMLheader .= '</a>';
					$HTMLheader .= print_user_picture($user,$vpl->get_course()->id,TRUE,FALSE,TRUE).' '.fullname($user);
				}
				$fg = $submission->get_submitted_fgm();
				$data = $fg->getFileData($filename);
			}
		}elseif($type == 2){

		}else{

		}
	}
}
$id = required_param('id', PARAM_INT);
$vpl = new mod_vpl($id);
$vpl->add_to_log('Diff', vpl_mod_href('similarity/diff.php','id',$id));
//Print header
vpl_sh_base::print_CSS();
$vpl->print_header(get_string('diff',VPL));
$vpl->print_view_tabs(basename(__FILE__));
//Get left file
vpl_diff::vpl_get_similfile('1',$HTMLheader1,$filename1,$data1);
//Get right file
vpl_diff::vpl_get_similfile('2',$HTMLheader2,$filename2,$data2);
//Show files
vpl_diff::show($filename1,$data1,$HTMLheader1,$filename2,$data2,$HTMLheader2);
$vpl->print_footer();

?>
