<?php
/**
 * @version		$Id: similarity_base.class.php,v 1.14 2010/01/29 13:03:13 juanca Exp $
 * @package		vpl. Similarity base and utility clases
 * @copyright	Copyright (C) 2009 Juan Carlos Rodríguez-del-Pino. All rights reserved.
 * @license		GNU/GPL, see LICENSE.txt or http://www.gnu.org/licenses/gpl-2.0.html
 * @author		Juan Carlos Rodriguez-del-Pino
 **/

require_once dirname(__FILE__).'/tokenizer_factory.class.php';

/**
 * @abstract class to represent files from any source 
 */
class vpl_file_from_base{
	public function show_info(){
	}
	public function can_access(){
		return false;
	}
}

/**
 * Represent information of a file from a directory
 */
class vpl_file_from_dir extends vpl_file_from_base{
	static $usersname=array();
	var $filename;
	var $userid;
	static public function add_user_name($userid,$username){
		if(!isset(self::$usersname[$userid])){
			self::$usersname[$userid]=$username;
		}
	}
	function __construct(&$filename,$userid=''){
		$this->filename=$filename;
		$this->userid=$userid;
	}
	public function show_info(){
		$ret='';
		$ret .='<a href="'.'file.php'.'">';
		$ret .=s($this->filename).' ';
		$ret .= '</a>';
		if($this->userid!=''){
			$ret .= ' '.self::$usersname[$this->userid];
		}
		return $ret;		
	}
}

/**
 * Represent information of a file from a zip file
 *
 */
class vpl_file_from_zipfile extends vpl_file_from_dir{
	function __construct(&$filename,$userid=''){
		parent::__construct($filename,$userid);
	}
	public function show_info(){
		$ret='';
		$ret .=s($this->filename);
		if($this->userid!=''){
			$ret .= ' '.self::$usersname[$this->userid];
		}
		return $ret;
	}
}

/**
 * Represent information of a file from other vpl activity
 *
 */
class vpl_file_from_activity extends vpl_file_from_base{
	static $vpls=array();
	var $vplid;
	var $filename;
	var $subid;
	var $userid;
	function __construct(&$filename,&$vpl,$submistance){
		$id = $vpl->get_instance()->id;
		if(!isset(self::$vpls[$id])){
			self::$vpls[$id]=$vpl;
		}
		$this->vplid = $id;
		$this->filename=$filename;
		$this->userid=$submistance->userid;
		$this->subid=$submistance->id;
	}
	public function show_info(){
		$vpl = self::$vpls[$this->vplid];
		$cmid = $vpl->get_course_module()->id;
		$ret='';
		if($this->userid>=0){
			$user = get_record('user','id',$this->userid);
		}else{
			$user = false;
		}
		if($user){
			$ret .='<a href="'.vpl_mod_href('/forms/submissionview.php','id',$cmid,'userid',$user->id).'">';
		}
		$ret .=s($this->filename).' ';
		if($user){
			$ret .= '</a>';
			$ret .= print_user_picture($user,$vpl->get_course()->id,TRUE,FALSE,TRUE).' '.fullname($user);
		}
		return $ret;		
	}
	public function can_access(){
		return true;
	}
	public function link_parms($t){
		return array('type'.$t=>1,'subid'.$t=>$this->subid,'filename'.$t => $this->filename);
	}
}


/**
 * Represent similarity preprocesing information of a file
 * from any source (directory, zip file or vpl activity)
 *
 */
class vpl_similarity_base{
	var $from;
	var $size;
	var $vecfrec;
	var $hashses;
	public function get_type(){
		return 0;
	}
	public function get_tokenizer(){
		return vpl_tokenizer_factory::get('base');
	}
	public function sintax_normalize(&$tokens){
		return $tokens;
	}
	public function init(&$data,&$from){
		$HASH_SIZE = 4;
		$HASH_REDUCTION = 1000;
		$this->from=$from;
		$this->size=0;
		$this->vecfrec = array();
		$this->hashes = array();
		//Get tokenizer for this language
		$tok = $this->get_tokenizer();
		//Parse code
		$tok->parse($data);
		//Normalize parsed data transforming or removing code
		$tokens = $this->sintax_normalize($tok->get_tokens());
		//Process tokens to get vector of frecuencies and size
		//Calculate hashes 
		//TODO improve performance using array of ints and numering operators
		$last = array();
		for($i=0; $i<$HASH_SIZE ; $i++){
			$last[$i] = '';
		}
		foreach($tokens as $token){
			if($token->type == vpl_token_type::operator){
				//Calculate hashes table
				for($i=0; $i<$HASH_SIZE-1 ; $i++){
					$last[$i] = $last[$i+1];
				}
				$last[$HASH_SIZE-1] = $token->value;
				$item = '';
				for($i=0; $i<$HASH_SIZE ; $i++){
					$item .= $last[$i];
				}
				$hash = crc32($item);
				if(isset($this->hashes[$hash%$HASH_REDUCTION])){
					$this->hashes[$hash%$HASH_REDUCTION]++;
				}else{
					$this->hashes[$hash%$HASH_REDUCTION] = 1;
				}
				$this->size++;
				if(isset($this->vecfrec[$token->value])){
					$this->vecfrec[$token->value]++;
				}else{
					$this->vecfrec[$token->value]=1;
				}
			}
		}
	}
	public function show_info(){
		return $this->from->show_info();
	}
	public function can_access(){
		return $this->from->can_access();
	}
	public function link_parms($t){
		return $this->from->link_parms($t);
	}
	
	
	/**
	 * Get similarity 1 among this file and other
	 * @param $other the other file info object
	 * @return number 0-100 %
	 */
	public function similarity1(&$other){
		$dif1=0;
		$taken=0;
		foreach($this->vecfrec as $op => $frec){
			if(isset($other->vecfrec[$op])){
				if($frec != $other->vecfrec[$op]){
					$dif1++;
				}
				$taken++;
			}else{
				$dif1++;
			}
		}
		$dif2 = count($other->vecfrec)-$taken;
		return 100*(1-(($dif1+$dif2)/(count($this->vecfrec)+count($other->vecfrec))));
	}

	/**
	 * Get similarity 2 among this file and other
	 * @param $other the other file info object
	 * @return number 0-100 %
	 */
	public function similarity2(&$other){
		$dif=0;
		$taken=0;
		foreach($this->vecfrec as $op => $frec){
			if(isset($other->vecfrec[$op])){
				$dif += abs($other->vecfrec[$op]-$frec);
				$taken +=$other->vecfrec[$op];
			}else{
				$dif+=$frec;
			}
		}
		$dif += $other->size - $taken;
		return 100*(1-($dif/($this->size+$other->size)));
	}

	/**
	 * Get similarity 3 among this file and other
	 * @param $other the other file info object
	 * @return number 0-100 %
	 */
	public function similarity3(&$other){
		$dif=0;
		$taken=0;
		foreach($this->hashes as $hash => $frec){
			if(isset($other->hashes[$hash])){
				$dif += abs($other->hashes[$hash]-$frec);
				$taken +=$other->hashes[$hash];
			}else{
				$dif+=$frec;
			}
		}
		$dif += $other->size - $taken;
		return 100*(1-($dif/($this->size+$other->size)));
	}
}

/**
 * Utility class to get list of preprocessed files
 *
 */
class vpl_similarity{
	static public function scan_activity(&$simil, $vpl, $filesselected=array()){
		global $CFG;
		$vpl->require_capability(VPL_SIMILARITY_CAPABILITY);
		$list = get_users_by_capability($vpl->get_context(), 'moodle/legacy:student', '', "u.lastname ASC", '', '', '', '', false);
		if($list == null){
			return;
		}
		$submissions = $vpl->all_last_user_submission();
		foreach ($list as $userinfo) {
			if(isset($submissions[$userinfo->id])){
				$origin = '';
				$subinstance = $submissions[$userinfo->id];
				$submission = new mod_vpl_submission($vpl,$subinstance);
				$subf = $submission->get_submitted_fgm();
				$filelist = $subf->getFileList();
				foreach($filelist as $filename){
					if(count($filesselected)>0 && !isset($filesselected[$filename])){
						continue;
					}
					$sim = vpl_similarity_factory::get($filename);
					if($sim){
						$data = $subf->getFileData($filename);
						$from =new vpl_file_from_activity($filename,$vpl,$subinstance);
						$sim->init($data,$from);
						if($sim->size>10){
							$simil[]=$sim;
						}
					}
				}
			}
		}
	}
	
	static public function scan_zip(&$simil, $zipfile, $vpl, $filesselected=array()){
		global $CFG;
		$ext = strtoupper(pathinfo($zipfile,PATHINFO_EXTENSION));
		if($ext != 'ZIP'){
			print_error('nozipfile');
		}
		//TODO change to new zip class when available
		//Get class to handle zip
		require_once($CFG->libdir . '/filelib.php');
		require_once($CFG->libdir . '/pclzip/pclzip.lib.php');
		$course = $vpl->get_course();
		$dirfilename=tempnam($CFG->dataroot . '/temp/'  , 'vpl_dirextract' );
		$destdir=$dirfilename.'.dir';
		mkdir($destdir,0700);
		$zipfilename=$CFG->dataroot.'/'.$course->id.'/'.$zipfile;
		debugging("Unzip file ".$zipfilename, DEBUG_DEVELOPER);
		$zip = new PclZip($zipfilename);
		$listfiles = $zip->extract(PCLZIP_OPT_PATH, $destdir);
		$toremove=strlen($destdir);
		foreach($listfiles as $file){
			if(!$file['folder']){
				//TODO search userifo
				$filename=$file['filename'];
				debugging("Examining file ".$filename, DEBUG_DEVELOPER);
				if(count($filesselected)>0 && !isset($filesselected[basename($filename)])){
					continue;
				}
				$sim = vpl_similarity_factory::get($filename);
				if($sim){
					$data = file_get_contents($filename);
					//TODO locate userid
					$from =new vpl_file_from_zipfile(substr($filename,$toremove));
					$sim->init($data,$from);
					if($sim->size>10){
						$simil[]=$sim;
					}
				}
			}
		}
		//remove temp dir
		unlink($dirfilename);
		vpl_delete_dir($destdir);	
	}
	static function get_dir_filelist($basepath,$dir){
        $filelist = array();
        $dirlist = array();
        if($dir > ''){
        	$basedir = $basepath.'/'.$dir;
        }else{
        	$basedir = $basepath;
        }
        if($dd = opendir($basedir)){
			while ($file=readdir($dd)) {
				if ($file[0] == '.'){
					continue;
				}
				$filepath=$dir.'/'.$file;
				if(is_file($basedir."/".$filepath)){
					$filelist[] = $filepath;
				}elseif(is_dir($basedir."/".$filepath)){
					$dirlist[] = $filepath;
				}
			}
			closedir($dd);
        }
        foreach($dirlist as $adir){
        	$otherlist = self::get_dir_filelist($basepath,$adir);
        	$filelist = array_merge($filelist,$otherlist);
        }
        return $filelist;
	}
	static public function scan_directory(&$simil, $dir, $filesselected=array()){
		$filelist = self::get_dir_filelist($dir,'');
		foreach($filelist as $file){
			$base = basename($file);
			if(count($filesselected)>0 && !isset($filesselected[$base])){
				continue;
			}
			$sim = vpl_similarity_factory::get($dir.'/'.$file);
			if($sim){
				$data = file_get_contents($file);
				//TODO locate userid
				$from =new vpl_file_from_dir($file);
				$sim->init($data,$from);
				if($sim->size>10){
					$simil[]=$sim;
				}
			}
		}
	}
	static public function get_order($s1,$s2=0,$s3=0){
		return ((int)$s1)*100000+((int)$s2)*100+(int)$s3;
	}
	
	static public function get_selected(&$files, $maxselected, $slimit){
		$vs1=array();
		$vs2=array();
		$vs3=array();
		$minlevel1=0;
		$minlevel2=0;
		$minlevel3=0;
		$selected = array();
		$jl=count($files);
		for($i=0; $i<$slimit; $i++){ //Search similarity with
			$current=&$files[$i];
			$current_type=$current->get_type();
			for($j=$i+1;$j<$jl; $j++){ //Compare with all others
				$other=&$files[$j];
				//If not the same language then skip
				if($current_type != $other->get_type()){
					continue;
				}
				//Calculate metrics
				$s1=$current->similarity1($other);
				$s2=$current->similarity2($other);
				$s3=$current->similarity3($other);
				if($s1>=$minlevel1 || $s2>=$minlevel2 || $s3 >= $minlevel3){
					$order = vpl_similarity::get_order($s1,$s2);
					$case = new StdClass();
					$case->first=&$files[$i];
					$case->second=&$files[$j];
					$case->s1=$s1;
					$case->s2=$s2;
					$case->s3=$s3;
					$case->comparation=(int)$s1.' / '.(int)$s2.' / '.(int)$s3.' / ';
					if($s1>=$minlevel1){
						$vs1[] = $case;
						if(count($vs1) > 2*$maxselected){
							self::filter_selected($vs1,$maxselected,$minlevel1,1);
						}
					}
					if($s2>=$minlevel2){
						$vs2[] = $case;
						if(count($vs2) > 2*$maxselected){
							self::filter_selected($vs2,$maxselected,$minlevel2,2);
						}
					}
					if($s3>=$minlevel3){
						$vs3[] = $case;
						if(count($vs3) > 2*$maxselected){
							self::filter_selected($vs3,$maxselected,$minlevel3,3);
						}
					}
				}
			}
		}
		self::filter_selected($vs1,$maxselected,$minlevel1,1);
		self::filter_selected($vs2,$maxselected,$minlevel2,2);
		self::filter_selected($vs3,$maxselected,$minlevel3,3);
		//Merge vs1, vs2 and vs3
		$max = $maxselected<count($vs1)? $maxselected : count($vs1);
		for($i=0; $i< $max ; $i++){
			$selected[]=$vs1[$i];
			$selected[]=$vs2[$i];
			$selected[]=$vs3[$i];
			$vs1[$i]->comparation .= '*';
			$vs2[$i]->comparation .= '*';
			$vs3[$i]->comparation .= '*';
		}
		$selected = array_unique($selected);
		return $selected;
	}

	static public function filter_selected(&$vec, $maxselected, &$minlevel,$sid){
		if(count($vec)>$maxselected){
			usort($vec,'vpl_similarity_base_cmp_selected'.$sid);
			$field = 's'.$sid;
			$minlevel=$vec[$maxselected-1]->$field;
			$chunkes=array_chunk($vec,$maxselected);
			$vec = $chunkes[0];
		}
	}
}

//uasort don't acept static public class member functions then define
//Standard functions

function vpl_similarity_base_cmp_selected1($a, $b){
	if ($a->s1 == $b->s1) {
        return 0;
   	}
    return ($a->s1 > $b->s1) ? -1 : 1;
}

function vpl_similarity_base_cmp_selected2($a, $b){
	if ($a->s2 == $b->s2) {
        return 0;
   	}
    return ($a->s2 > $b->s2) ? -1 : 1;
}

function vpl_similarity_base_cmp_selected3($a, $b){
	if ($a->s3 == $b->s3) {
        return 0;
   	}
    return ($a->s3 > $b->s3) ? -1 : 1;
}
?>